;; Generated package description from apache-mode.el  -*- no-byte-compile: t -*-
(define-package "apache-mode" "2.2.0.0.20251120.174624" "Major mode for editing Apache httpd configuration files" 'nil :commit "a3db91fc6c4d46ca0625e79df5d6689a607b1927" :authors '(("Karl Chen" . "quarl@nospam.quarl.org")) :maintainer '("USAMI Kenta" . "tadsan@zonu.me") :keywords '("languages" "faces") :url "https://github.com/emacs-php/apache-mode")
